#include <stdio.h>
#include <string.h>
#include "common.h"

#define FILETYPE_PLAIN 0
#define FILETYPE_INI   1
char *get_embedded_script(char *filename, int startLine) {
	FILE *file;
	char *buf, *pbuf;
	BOOL lineComment = FALSE, quoted = FALSE, newCommand = TRUE;
	int line = 1, brackets = 0, ch, lastch = 0, size = 4096, fileType = FILETYPE_PLAIN;

	/* Check for INI file */
	if (strnicmp(filename + strlen(filename) - 4, ".ini", 4) == 0) {
		fileType = FILETYPE_INI;
	}

	/* Open file */
	file = fopen(filename, "r");
	if (file == NULL) {
		return NULL;
	}

	if ((buf = (char *)malloc(size)) == NULL) {
		return (char *)-1;
	}
	pbuf = buf;
	while ((ch = fgetc(file)) != EOF) {
		if (line > startLine) { /* Start parsing code */

			/* Copy the script data into the buffer */
			*pbuf++ = ch;
			if (pbuf >= buf + size - 1) { /* Give buffer more space if needed */
				size *= 2;
				if ((buf = (char *)realloc(buf, size)) == NULL) {
					return NULL;
				}
			}

			if (lineComment) { /* Inside a # comment */
				if (ch == '\n') { lineComment = FALSE; newCommand = TRUE; }
			}
			else if (quoted) { /* Inside a quote */
				if (ch == '"' && lastch != '\\') quoted = FALSE; 
			}
			else {
				/* The reason we're here, counting brackets */
				if (ch == '{') brackets++;
				if (ch == '}') brackets--;
				if (brackets < 0) break;

				if (lastch != '\\') {
					/* Check for a new command */
					if (ch == ';' || ch == '\n') newCommand = TRUE; 
					else if (ch != ' ' && ch != '\t') newCommand = FALSE;

					/* Entering a # comment */
					if (newCommand && ch == '#') lineComment = TRUE;

					/* Entering a quote */
					if (ch == '"') quoted = TRUE;
				}
			}

			/* Need to strip nNNN= from the beginning of lines in INI files */
			if (fileType == FILETYPE_INI) {
				if (ch == '\n') {
					int ic;
					while (ic = fgetc(file)) {
						if (ic == '=') break;
					}
				}
			}

			lastch = ch;
		}

		if (ch == '\n') line++; 
	}
	*(--pbuf) = 0;

	fclose(file);

	return buf;
}

Tcl_HashTable code_cache;

/* Builds the procedure name in the form "file:line" */
Tcl_Obj *BuildProcName(Tcl_Obj *filename, Tcl_Obj *line) {
	Tcl_Obj *procName = Tcl_DuplicateObj(filename);
	Tcl_AppendToObj(procName, ":", 1);
	Tcl_AppendObjToObj(procName, line);
	return procName;
}

int mIRC_Execute(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) {
	int len, line, cache = FALSE;
	Tcl_Obj *result, *procName;
	char *script, *filename;

	if (objc < 3) {
		Tcl_WrongNumArgs(interp, objc, objv, "mirc_embed filename line ?-cache?");
		return TCL_ERROR;
	}
	if (objc == 4) { /* If a 3rd parameter == -cache, cache is true */
		cache = Tcl_StringMatch(Tcl_GetStringFromObj(objv[3], &len), "-cache");
	}

	filename = Tcl_GetStringFromObj(objv[1], &len);
	
	if ((Tcl_GetIntFromObj(interp, objv[2], &line)) == TCL_ERROR) {
		/* Problem parsing line number */
		result = Tcl_GetObjResult(interp);
		Tcl_SetStringObj(result, "invalid line number", -1);
		return TCL_ERROR;
	}

	/* (If cache is on) check if we have this code in the cache */
	if (cache) {
		Tcl_HashEntry *entry;

		/* Build the function name */
		procName = BuildProcName(objv[1], objv[2]);

		if (entry = Tcl_FindHashEntry(&code_cache, (const char *)procName)) { 
			/* We have it. Execute it and we're done */
			Tcl_Obj *scriptObj = (Tcl_Obj *)Tcl_GetHashValue(entry);
			return Tcl_EvalObjEx(interp, scriptObj, TCL_EVAL_GLOBAL);
		}
	}

	if ((script = get_embedded_script(filename, line)) > 0) {
		/* Successfully parsed script */
		int retval = TCL_OK;
		
		if (cache) { /* Add the script to the cache and execute it */
			Tcl_Obj *scriptObj = Tcl_NewStringObj(script, -1);
			Tcl_HashEntry *entry = Tcl_CreateHashEntry(&code_cache, (const char *)procName, (ClientData)scriptObj);

			/* BUGFIX: Tcl_CreateHashEntry does not properly set clientData (?) */
			entry->clientData = (ClientData)scriptObj; 

			retval = Tcl_EvalObjEx(interp, scriptObj, TCL_EVAL_GLOBAL);
		}
		else { /* No caching, execute it without byte compilation */
			retval = Tcl_EvalEx(interp, script, -1, TCL_EVAL_GLOBAL);
		}

		free(script);
		return retval;
	}
	else {
		result = Tcl_GetObjResult(interp);

		if (script == (char *)-1) { /* File not found */
			Tcl_SetStringObj(result, "could not find file '", -1);
			Tcl_AppendObjToObj(result, objv[1]);
			Tcl_AppendToObj(result, "'", -1);
		}
		else { /* Problem with memory */
			Tcl_SetStringObj(result, "out of memory while parsing script '", -1);
			Tcl_AppendObjToObj(result, objv[1]);
			Tcl_AppendToObj(result, "' at line ", -1);
			Tcl_AppendObjToObj(result, objv[2]);
		}
		return TCL_ERROR;
	}

	return TCL_ERROR;
}

/* Initialization */

int Embed_Init(Tcl_Interp *interp) {
	Tcl_InitObjHashTable(&code_cache);
	Tcl_CreateObjCommand(interp, "mirc_embed", mIRC_Execute, NULL, NULL);
	
	return TCL_OK;
}